---
comment:  This document is in 'markdown' format.  To get a pdf, type
  $ pandoc -o /tmp/doc.pdf doc.md
---

### Reality check progress report, Sept. 9 2024

As part of IRCAM's participation in DAFNE+, a continuous integration platform named Reality Check is under development with the aim of setting up an automatic testing system for realizations of live computer music.  The great majority of such realizations use one of Pure Data, Cycling74 Max, or Supercollider.   Of these, Max is by far the most prevalent at IRCAM (and probably elsewhere as well), whereas Pure Data is less frequently used but lends itself more readily to continuous integration; these two are the focus of this work.

### Pure Data

Version 0.03 of Reality Check for Pd was released in March 2024 along with two test realizations by Philippe Manoury (an IRCAM production) and Kerry Hagan (University of Illinois).  A third realization by Rand Steiger (UCSD) has been prepared and tested but not yet released.  Between then, these realizations cover much of Pure Data's built-in functionality and test the possibility of including external libraries whose sources may be recompiled as part of a testing script.  We can thus assure CI verification of self-contained Pure Data realizations.

In the intervening months two new capabilities have been added and tested: third-party VST plug-ins (using Synful as an example) and the Antescofo scripting language.  In both cases it was necessary to track down and correct sources of non-determinism.  Antescofo in particular is used almost universally in IRCAM's new productions, since it offers functionalities that complement those of Max and Pure Data.

The combination of Pure Data, Synful, and Antescofo is being put to a severe test by an upcoming opera by Philippe Manoury.  At present it is possible to run an entire scene of the opera and verify its bitwise reproducibility from run to run.

### Max

An early prototype of a Max variant of Reality Check is in an early testing stage.  Although not ready for production work, the prototype shows that it is feasible although difficult to adapt Max patches into a CI environment.

As a commercial, closed-source software environment, Max does not lend itself as easily to CI as does Pure Data.  Over a span of decades Max has seen the introduction of many new features along with the disappearance or incompatible evolution of old ones.  Moreover, in its usual mode of operation Max is non-deterministic, and it is as yet uncertain whether it is possible at all to obtain exactly repeatable results from a Max run.

On the other hand Max is far more widely used than Pure Data, partly for historical reasons but especially because it is much more feature-rich and user friendly.  It is understandable that a composer, faced with a hard deadline and limited studio time, would reach for Max as the most expedient solution available.

Whereas the Pd version of Reality Check uses the pd~ object to virtualize the entire realization in its own instance of Pd, this has not proved as convenient in Max.  There is indeed a Max~ object which aims to do the same thing but its I/O capabilities are limited.  Audio input and output don't work in the contained patch and so it must be adapted to use inlets and outlets.  Similarly, messages going into and out of into a max~ sub-process must be formatted as MIDI.  It was judged more expedient to have Reality Check include the realization to be tested as a Max abstraction.  This necessitates adapting the realization to offer inlets and outlets for all control and audio I/O, as would have been necessary anyway in Max.

The prototype takes advantage of a Max scheduler option, NonRealTime, that substantially limits the variability of Max's behavior from run to run.  In practice it appears to be possible to get timing variations down to about a millisecond.  (In theory it is exact but not so far in practice).

Since the sound output may be subject to small differences in timing it is necessary to devise a more flexible test of equivalence than exact equality.  To this end a test has been implemented that compares short-time magnitude spectra, which should not change appreciably with small time variations.  Specifically, a Pd patch, "compare", has been developed that evaluates the global maxima and minima of the spectral deviation defined as:

${{|Y[k]| + \epsilon} \over {|X[k]| + \epsilon}}$

where $X[k]$ and $Y[k]$ are the short-time windowed Fourier transforms (WSTFTs) of two sounds to be compared, and $\epsilon$ is a small number chosen to suppress differences in inaudibly quiet frequency bands.  In preliminary tests this discriminator is able to detect differences in amplitude or in timing comparable to the WSTFT window size but not to the small timing differences encountered in Max in non-real-time mode.

### bullet points

- Pure data version tested on a major upcoming production

- integration of antescofo language

- preliminary tests of Max version of Reality check

- built and tested a discriminator that accepts small timing errors




